/*
 * Copyright 2016 Steve Nickolas, except as designated below.
 * Other copyrights may apply to plugin code.
 * 
 * CRC32 functionality in modapple is based on code copyright (C) 1986 by Gary 
 * S. Brown.  You may use this code, or tables extracted from it, as desired 
 * without restriction.  This applies to the table crc32_tab and the function 
 * crc32() in modapple.c.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal with the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 *   1. Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimers.
 *   2. Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimers in the 
 *      documentation and/or other materials provided with the distribution.
 *   3. Neither the name of the author, nor the names of its contributors, may 
 *      be used to endorse or promote products derived from this Software 
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

static char *verstr=
  "MoDapple  Version 0.33\n"
  "Copyright 2016 Steve Nickolas\n"
  "Other copyrights apply to plugin code\n";
static char *ESTR_NOTMOD ="%s is not a module\n",
            *ESTR_NOFUNC ="Could not find a '%s' function in %s\n",
            *ESTR_FSIZE  ="%s: Wrong file size (%lu expected, %lu found)\n",
            *ESTR_SHORT  ="%s: Short read (%lu expected, %lu found\n",
            *ESTR_INITMOD="Could not initialize module %s\n",
            *ESTR_SLOT   ="Slot %d",
            *ESTR_IS     =" is %s Version %s\n%s\n",
            *ESTR_LINK   ="Could not link %s to slot %d\n",
            *ESTR_NOTTY  ="No console loaded\n",
            *ESTR_NOCPU  ="No CPU loaded\n",
            *ESTR_NOBIOS ="No BIOS loaded\n",
            *ESTR_BADSLOT="Invalid slot %d\n",
            *ESTR_OPEN   ="Could not open module %s\n",
            *ESTR_ISNOT  ="%s is not a ",
            *ESTR_NOTSLOT="slot module\n",
            *ESTR_NOTTTY ="console module\n",
            *ESTR_NOTCPU ="CPU module\n",
            *ESTR_CPU    ="CPU",
            *ESTR_MULT   ="CPU Speed is %d MHz\n",
            *ESTR_TTY    ="Console",
            *ESTR_SND1   ="Could not open sound device\n",
            *ESTR_SND2   ="Could not prepare sound device\n",
            *ESTR_SND3   ="Successfully initialized sound device\n",
            *ESTR_SND4   ="Waiting for sound buffer to flush\n";

static char *STRING_UNDERSCORE_STRING="%s_%s";
static char *MOD_TYPE="mod_type",
            *MOD_NAME="mod_name",
            *MOD_PREF="mod_pref",
            *MOD_VERS="mod_vers",
            *MOD_COPR="mod_copr",
            *INI_SYSTEM   ="System",
            *INI_SLOTCARDS="SlotCards",
            *INI_BIOS     ="BIOS",
            *INI_CONSOLE  ="Console",
            *INI_CPU      ="CPU",
            *INI_MULT     ="CPUMultiplier",
            *INI_DEF_ROM  ="apple2p.rom";

static char *ARBY="rb"; /* for reading files */

HWAVEOUT hwaveout;
WAVEHDR hdr;
uint8_t sndbuf[157200];
uint32_t sndbufptr;
uint8_t sndon;
uint8_t polarity, amplitude;

int openwav (void)
{
 WAVEFORMATEX fmt;
 
 fmt.cbSize=0;
 fmt.nChannels=1;
 fmt.nSamplesPerSec=(15720);
 fmt.wBitsPerSample=8;
 fmt.nBlockAlign=(fmt.nChannels*fmt.wBitsPerSample)>>3;
 fmt.nAvgBytesPerSec=fmt.nSamplesPerSec*fmt.nBlockAlign;
 fmt.wFormatTag=WAVE_FORMAT_PCM;
 return waveOutOpen(&hwaveout,WAVE_MAPPER,&fmt,0,0,CALLBACK_NULL);
}

typedef struct
{
 HMODULE  dll;
 uint32_t *mod_type;
 uint8_t  **mod_name;
 uint8_t  **mod_pref;
 uint8_t  **mod_vers;
 uint8_t  **mod_copr;
 uint8_t  **preload_name;
 uint16_t *preload_size;
 uint8_t  **preload_ptr;
 void     (*write16)     (uint16_t a, uint8_t v);
 void     (*write256)    (uint16_t a, uint8_t v);
 void     (*write2048)   (uint16_t a, uint8_t v);
 uint8_t  (*read16)      (uint16_t a);
 uint8_t  (*read256)     (uint16_t a);
 uint8_t  (*read2048)    (uint16_t a);
 int16_t  (*override_rd) (uint16_t a);
 int16_t  (*override_wr) (uint16_t a, uint8_t v);
 int      (*reset)       (void);
 int      (*loop)        (int cycles);
 
 /*
  * Formerly, init() was a boolean.
  * For plugins intended for MoDapple 0.09 and later, if init() returns "6502"
  * then it needs pointers to cpuread() and cpuwrite().  The function to take
  * these pointers is only linked if it is required.
  */
 int      (*init)        (uint8_t s);
 int      (*deinit)      (void);
 
 void     (*dmainit)     (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t));

 void     (*config)      (void);
} SLOT;

SLOT slot[8];

HINSTANCE hinstance;

typedef struct {
 uint8_t iiemode; /* emulating a //e */
 uint8_t store80,
         rdaux,
         wraux,
         lcaux,
         cxoverride,
         c3override,
         txt80,
         dblgrx,
         mostxt,
         iou,
         an3,
         alt[2];
 uint8_t lcread,
         lcwrite,
         lcalt;
 uint8_t nogx,
         split,
         page2,
         hires;

 uint8_t *ram;
 uint8_t enh;
} IIE;

IIE iie;

int cpumultiplier;

uint8_t  ram[131072], rom[16384];

uint8_t  fnstr[256];

/* Data pointers and functinos for TTY driver */
uint32_t *tty_mod_type;
uint8_t  **tty_mod_name;
uint8_t  **tty_mod_pref;
uint8_t  **tty_mod_vers;
uint8_t  **tty_mod_copr;
uint8_t  *ttykey;
void     (*ttywrite)   (uint32_t a, uint8_t v);
void     (*ttyiie)     (void);
void     (*ttyout)     (uint8_t a, uint8_t v);
uint8_t  (*ttyin)      (uint8_t a);
void     (*ttydeinit)  (void);
void     (*ttyinit)    (IIE*);
int      (*ttyloop)    (void);

/* Data pointers and functinos for CPU driver */
uint32_t *cpu_mod_type;
uint8_t  **cpu_mod_name;
uint8_t  **cpu_mod_pref;
uint8_t  **cpu_mod_vers;
uint8_t  **cpu_mod_copr;
void     (*cpureset)     (void);
void     (*cpurun)       (void);
void     (*cpuinit)      (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t), int8_t (*v_cpuloop)(uint32_t));
void     (*cpuseta)      (uint8_t val);
void     (*cpusetx)      (uint8_t val);
void     (*cpusety)      (uint8_t val);
void     (*cpusets)      (uint8_t val);
void     (*cpusetp)      (uint8_t val);
void     (*cpusetpc)     (uint16_t val);
void     (*cpusetrate)   (int val);
uint8_t  (*cpugeta)      (void);
uint8_t  (*cpugetx)      (void);
uint8_t  (*cpugety)      (void);
uint8_t  (*cpugets)      (void);
uint8_t  (*cpugetp)      (void);
uint16_t (*cpugetpc)     (void);
int      (*cpugetrate)   (void);
int      (*cpugetcycles) (void);

HMODULE mod_tty, mod_cpu;

void canlibs (void)
{
 if (mod_tty) FreeLibrary(mod_tty);
 if (mod_cpu) FreeLibrary(mod_cpu);
 
 if (sndon)
 {
  printf (ESTR_SND4);
  while (waveOutUnprepareHeader(hwaveout,&hdr,sizeof(hdr))==WAVERR_STILLPLAYING) Sleep(100);
  waveOutClose(hwaveout);
 }
}

void notamod (char *filename)
{
 canlibs();
 fprintf (stderr, ESTR_ISNOT, filename);
 fprintf (stderr, ESTR_NOTMOD);
 exit(1);
}

void nofunc (char *filename, char *ident)
{
 canlibs();
 fprintf (stderr, ESTR_NOFUNC, ident, filename);
 exit(1);
}

int nofunc2 (int slotno, char *filename, char *ident)
{
 FreeLibrary(slot[slotno].dll);
 slot[slotno].dll=0;
 fprintf (stderr, ESTR_NOFUNC, ident, filename);
 return -1;
}

#define TFUNCTINO(a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*tty_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_tty,fnstr))) nofunc(ttymod, b)
#define CFUNCTINO(a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*cpu_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_cpu,fnstr))) nofunc(cpumod, b)
#define SFUNCTINO(s,a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*(slot[s].mod_pref),b); if (!(slot[s].a=(void*)GetProcAddress(slot[s].dll,fnstr))) return nofunc2(s, module, b)

static uint32_t crc32_tab[] = {
     0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f,
     0xe963a535, 0x9e6495a3,  0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
     0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064, 0x6ab020f2,
     0xf3b97148, 0x84be41de,  0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
     0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,   0x14015c4f, 0x63066cd9,
     0xfa0f3d63, 0x8d080df5,  0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
     0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,   0x35b5a8fa, 0x42b2986c,
     0xdbbbc9d6, 0xacbcf940,  0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
     0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423,
     0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
     0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,   0x76dc4190, 0x01db7106,
     0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
     0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d,
     0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
     0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
     0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
     0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7,
     0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
     0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa,
     0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
     0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81,
     0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
     0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12, 0x94643b84,
     0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
     0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
     0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
     0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8, 0xa1d1937e,
     0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
     0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55,
     0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
     0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28,
     0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
     0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f,
     0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
     0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
     0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
     0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69,
     0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
     0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc,
     0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
     0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693,
     0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
     0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

uint32_t
crc32(uint32_t crc, const void *buf, size_t size)
{
     const uint8_t *p;

     p = buf;
     crc = crc ^ ~0U;

     while (size--)
          crc = crc32_tab[(crc ^ *p++) & 0xFF] ^ (crc >> 8);

     return crc ^ ~0U;
}

uint32_t crcfile (char *filename)
{
 long l;
 uint32_t roll;
 FILE *file;
 
 uint8_t buf[65536];
 file=fopen(filename,"rb");
 if (!file)
 {
  perror(filename);
  return 0;
 }
 roll=0;
 fseek(file,0,SEEK_END);
 l=ftell(file);
 fseek(file,0,SEEK_SET);
 while (l>0x0000FFFF)
 {
  fread(buf,1,65536,file);
  roll=crc32(roll,buf,65536);
  l-=65536;
 }
 fread(buf,1,l,file);
 roll=crc32(roll,buf,l);
 return roll;
}

int bload (char *filename, int expect, uint8_t *pointer, uint8_t quiet)
{
 FILE *file;
 long l;
 
 file=fopen(filename,ARBY);
 if (!file)
 {
  perror(filename);
  return -1;
 }
 
 fseek(file,0,SEEK_END);
 l=ftell(file);
 fseek(file,0,SEEK_SET);
 if (l!=expect)
 {
  fclose(file);
  if (!quiet) fprintf (stderr, ESTR_FSIZE, filename, expect, l);
  return -1;
 }
 l=fread(pointer,1,expect,file);
 fclose(file);
 if (l<expect)
 {
  fprintf(stderr, ESTR_SHORT, filename, expect, l);
  return -1;
 }
 
 return 0;
}

/*****************************************************************************/
/*
 * Start of a joystick driver.  XXX: This code should be in a DLL.
 */

unsigned int GameMinX = 26, GameMinY = 26;
unsigned int GameMaxX = 27, GameMaxY = 27;
unsigned long ClockX, ClockY;                /* Timeouts for 556 timer */
unsigned int LeftAltDown = 0, RightAltDown = 0;

static JOYINFO joyinfo;
static unsigned int nojoys, whichjoy;
static BOOL idjoy1, idjoy2;
static int gotstick;

void joyinit (void)
{
 gotstick=0;
 
 if (!(nojoys=joyGetNumDevs())) return;
 
 idjoy1=(joyGetPos(JOYSTICKID1,&joyinfo)!=JOYERR_UNPLUGGED);
 idjoy2=((nojoys==2)&&(joyGetPos(JOYSTICKID2,&joyinfo)!=JOYERR_UNPLUGGED));
 if (idjoy1||idjoy2)
 {
  gotstick=1;
  whichjoy=idjoy1?JOYSTICKID1:JOYSTICKID2;
 }
}

uint16_t jx, jy, jc;

unsigned char ResetGameTimer( unsigned short Address )
{
 joyGetPos(whichjoy,&joyinfo);
 jx=joyinfo.wXpos;
 jy=joyinfo.wYpos;
 jc=cpugetcycles();
}

unsigned char ReadGameTimer( unsigned short Address )
{
 int32_t jt;
 uint16_t axis;
 
 axis=(Address&1)?jy:jx;
 
 jt=cpugetcycles()-jc;
 //printf ("Timer: %08X  Axis: %04X\n", jt, axis);
 return (((int16_t)jt-(axis>>8))<0)?0x80:0x00; /* this is wrong, but how to fix? */
}

/*****************************************************************************/

static int fast6, fast6on;

static int holdc8;

/*
 * The ][+'s language card is a plugin, so that it can be added and removed at
 * will.  The //e and later MUST have this feature, and so a second language
 * card is hardwired into the shell for this support.
 */

uint8_t lclast;

static void iielc (uint16_t a)
{
 switch (a&0x0F)
 {
  case 0x0:
  case 0x4:
   iie.lcalt=0x80;
   iie.lcread=0x80;
   iie.lcwrite=0x00;
   break;
  case 0x1:
  case 0x5:
   iie.lcalt=0x80;
   iie.lcread=0x00;
   iie.lcwrite=0x80;
   break;
  case 0x2:
  case 0x6:
   iie.lcalt=0x80;
   iie.lcread=0x00;
   iie.lcwrite=0x00;
   break;
  case 0x3:
  case 0x7:
   iie.lcalt=0x80;
   iie.lcread=0x80;
   iie.lcwrite=0x80;
   break;
   
  case 0x8:
  case 0xC:
   iie.lcalt=0x00;
   iie.lcread=0x80;
   iie.lcwrite=0x00;
   break;
  case 0x9:
  case 0xD:
   iie.lcalt=0x00;
   iie.lcread=0x00;
   iie.lcwrite=0x80;
   break;
  case 0xA:
  case 0xE:
   iie.lcalt=0x00;
   iie.lcread=0x00;
   iie.lcwrite=0x00;
   break;
  case 0xB:
  case 0xF:
   iie.lcalt=0x00;
   iie.lcread=0x80;
   iie.lcwrite=0x80;
   break;
 }
 
 //if (lclast!=(a&0x0F)) iie.lcwrite=0x00;
 //lclast=(a&0x0F);
}

void opt_ttywrite (uint32_t a, uint8_t v)
{
 if (ram[a]!=v)
 {
  ram[a]=v;
  ttywrite(a,v);
 }
}

void cpuwrite(uint16_t a, uint8_t v)
{
 int t, o;
 
 if ((iie.iiemode)&&((a&0xFFF0)==0xC000))
 {
  switch (a&0x000F)
  {
   case 0x0:
    iie.store80=0x00;
    break;
   case 0x1:
    iie.store80=0x80;
    break;
   case 0x2:
    iie.rdaux=0x00;
    break;
   case 0x3:
    iie.rdaux=0x80;
    break;
   case 0x4:
    iie.wraux=0x00;
    break;
   case 0x5:
    iie.wraux=0x80;
    break;
   case 0x6:
    iie.cxoverride=0x00;
    break;
   case 0x7:
    iie.cxoverride=0x80;
    break;
   case 0x8:
    iie.lcaux=0x00;
    break;
   case 0x9:
    iie.lcaux=0x80;
    break;
   case 0xA:
    iie.c3override=0x00;
    break;
   case 0xB:
    iie.c3override=0x80;
    break;
   case 0xC:
    iie.txt80=0x00;
    break;
   case 0xD:
    iie.txt80=0x80;
    break;
   case 0xE:
    iie.mostxt=0x00;
    break;
   case 0xF:
    iie.mostxt=0x80;
    break;
  }
  ttyiie();
  return;
 }

 if ((a==0xC0E8)&&(fast6)) fast6on=0;
 if ((a==0xC0E9)&&(fast6)) fast6on=1;

 for (t=7; t>=(iie.iiemode?1:0); t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_wr(a, v);
   if (!o) return;
  }
 }
 
 if ((iie.iiemode)&&((a&0xFFF0)==0xC080))
 {
  iielc(a&0xF); 
  return;  
 }
 
 if (a<0x0200)
 {
  if ((iie.iiemode)&&(iie.lcaux)) ram[65536+a]=v; else ram[a]=v;
  return;
 }
 
 if ((iie.iiemode)&&(iie.store80))
 {
  if ((a>1023)&&(a<2048))               {opt_ttywrite((iie.page2?65536:0)|a,v); return;}
  if ((a>8192)&&(a<16384)&&(iie.hires)) {opt_ttywrite((iie.page2?65536:0)|a,v); return;}
 }
 else
 {
  if ((a>1023)&&(a<2048))  
  {
   opt_ttywrite((iie.wraux?65536:0)|a,v); 
   return;
  }
  if ((a>8191)&&(a<16384)) 
  {
   opt_ttywrite((iie.wraux?65536:0)|a,v); 
   return;
  }
 }
 if ((a>2047)&&(a<3072))   
 {
  opt_ttywrite((iie.wraux?65536:0)|a,v); 
  return;
 }
 if ((a>16383)&&(a<24576)) 
 {
  opt_ttywrite((iie.wraux?65536:0)|a,v); 
  return;
 }
 
 if (a<49152)
 {
  ram[(((iie.iiemode)&&(iie.wraux))?65536:0)|a]=v;
  return;
 }
 
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   slot[s].write16(a,v);
   return;
  }
 }
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  s=a&0x0700;
  s>>=8;
  holdc8=s;
  if (slot[s].dll)
  {
   slot[s].write256(a,v);
   return;
  }
 }

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  if ((iie.iiemode)&&(iie.cxoverride)) return;
  if (!slot[holdc8].dll) return;
  slot[holdc8].write2048(a-0xC800,v);
  if (a==0xCFFF) holdc8=0;
  return;
 }

 if (a>=0xD000)
 {
  if ((iie.iiemode)&&(iie.lcwrite))
  {
   if (a<0xE000)
   {
    if (iie.lcalt)
    {
     if (iie.lcaux) {ram[65536+(a-0x1000)]=v; return;}
     
     ram[a-0x1000]=v;
     return;
    }
   }
   if (iie.lcaux) {ram[65536+a]=v; return;}
   ram[a]=v;
  }
  return;
 }
 
 if (iie.iiemode)
 {
  if (a==0xC05E) {iie.dblgrx=0x80; if (!iie.iou) iie.an3=0x00; return;}
  if (a==0xC05F) {iie.dblgrx=0x00; if (!iie.iou) iie.an3=0x80; return;}
  if (a==0xC07E) {iie.iou=0x00; return;}
  if (a==0xC07F) {iie.iou=0x80; return;}
 }
 
 if ((a&0xFFF0)==0xC000)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC010)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC030)  {polarity=!polarity; return;}
 if ((a&0xFFF0)==0xC050)  {ttyout(a,v); return;}
 
 if (a==0xC070) {ResetGameTimer(a); return;}
}

/* User level scanline; XXX might not be in synch with the video plugin */
uint16_t ulscanline;

uint8_t cpuread(uint16_t a)
{
 int t, o;
 
 if ((a==0xC0E8)&&(fast6)) fast6on=0;
 if ((a==0xC0E9)&&(fast6)) fast6on=1;

 for (t=7; t>=(iie.iiemode?1:0); t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_rd(a);
   if (o>=0) 
   {
    return o;
   }
  }
 }

 if ((iie.iiemode)&&((a&0xFFF0)==0xC080)) 
 {
  iielc(a&0xF); 
  return 0;
 }

 if (a<0x0200)
 {
  if ((iie.iiemode)&&(iie.lcaux)) return ram[65536+a];
  return ram[a];
 }

 /* previously we used a ttyread() to get the info, but we shadow it here,
    so why waste the function call? */
 if ((iie.iiemode)&&(iie.store80))
 {
  if (iie.page2)
  {
   if ((a>1023)&&(a<2048))               return ram[65536+a];
   if ((a>8191)&&(a<16384)&&(iie.hires)) return ram[65536+a];
  }
  else
  {
   if ((a>1023)&&(a<2048))               return ram[a];
   if ((a>8191)&&(a<16384)&&(iie.hires)) return ram[a];
  }
 }
 
 if ((iie.iiemode)&&(iie.rdaux)&&(a<49152)) return ram[65536+a];
 if (a<49152)                               return ram[a];
 
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   return slot[s].read16(a);
  }
  else return 0;
 }
 
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  
  s=a&0x0700;
  s>>=8;

  holdc8=s;
  if ((iie.iiemode)&&(iie.cxoverride)) return rom[a&0x0FFF];
  if (s==3)
  {
   if ((iie.iiemode)&&(!iie.c3override)) return rom[a&0x0FFF];
  }

  if (slot[s].dll)
  {
   return slot[s].read256(a);
  }
  return rand()&0x80; /* well, it does the job */
 }

 if (((a&0xF800)==0xC800)&&(iie.iiemode)&&(iie.cxoverride)) return rom[a-0xC000];

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  uint8_t c;
  
  if (holdc8==3)
   if (iie.iiemode) return rom[a&0x0FFF];

  if (!slot[holdc8].dll) return 255;
  c=slot[holdc8].read2048(a-0xC800);
  if (a==0xCFFF) holdc8=0;
  return c;
 }

 if (a>=0xD000)
 {
  if (iie.iiemode)
  {
   if (iie.lcread)
   {
    if (a<0xE000)
    {
     if (iie.lcalt)
     {
      if (iie.lcaux) return ram[65536+(a-0x1000)];
      return ram[a-0x1000];
     }
    }
    if (iie.lcaux) return ram[65536+a];
    return ram[a];
   }
   return rom[a-0xC000];
  }
  return rom[a-0xD000];
 }
 
 if ((a&0xFFF0)==0xC000)  return ttyin(a);
 if ((a&0xFFF0)==0xC010)
 {
  if (iie.iiemode)
  {
   uint8_t k;
   
   if (!(a&0x0F)) k=ttyin(a)&0x7F; else k=ttyin(0)&0x7F;
   switch (a&0x0F)
   {
    case 0x0:
     break;
    case 0x1:
     k|=iie.lcalt;
     break;
    case 0x2:
     k|=iie.lcread;
     break;
    case 0x3:
     k|=iie.rdaux;
     break;
    case 0x4:
     k|=iie.wraux;
     break;
    case 0x5:
     k|=iie.cxoverride;
     break;
    case 0x6:
     k|=iie.lcaux;
     break;
    case 0x7:
     k|=iie.c3override;
     break;
    case 0x8:
     k|=iie.store80;
     break;
    case 0x9:
     k|=(ulscanline<192)?0x80:0x00;
     break;
    case 0xA:
     k|=(iie.nogx?0x80:0x00);
     break;
    case 0xB:
     k|=(iie.split?0x80:0x00);
     break;
    case 0xC:
     k|=(iie.page2?0x80:0x00);
     break;
    case 0xD:
     k|=(iie.hires?0x80:0x00);
     break;
    case 0xE:
     k|=iie.mostxt;
     break;
    case 0xF:
     k|=iie.txt80;
     break;
   }
   return k;
  }
  return ttyin(a);
 }

 if (iie.iiemode)
 {
  if (a==0xC05E) {iie.dblgrx=0x80; if (!iie.iou) iie.an3=0x00; return 0;}
  if (a==0xC05F) {iie.dblgrx=0x00; if (!iie.iou) iie.an3=0x80; return 0;}

  if (a==0xC07E) {return iie.iou;}
  if (a==0xC07F) {return iie.dblgrx;}
 }

 if (a==0xC070) {ResetGameTimer(a); return 0;}

 joyGetPos(whichjoy,&joyinfo);

 if ((a==0xC061)||(a==0xC069))
 {
  uint8_t b;
  
  if (gotstick) b=(joyinfo.wButtons&JOY_BUTTON1)?0x80:0x00;
  if (iie.iiemode) b|=iie.alt[0];
  return b;
 }
 if ((a==0xC062)||(a==0xC06A))
 {
  uint8_t b;
  
  if (gotstick) b=(joyinfo.wButtons&JOY_BUTTON2)?0x80:0x00;
  if (iie.iiemode) b|=iie.alt[1];
  return b;
 }
 if ((a==0xC064)||(a==0xC065)||(a==0xC06C)||(a==0xC06D))
 {
  return ReadGameTimer(a&0xFF);
 }

 if ((a&0xFFF0)==0xC030)  {polarity=!polarity; return 255;}
 if ((a&0xFFF0)==0xC050)  return ttyin(a);

 return 0;
}

static int die;

void reset_all(void)
{
 int t;
 
 fast6on=0;
 
 ulscanline=0;
 for (t=7; t>=0; t--)
 {
  if (slot[t].dll)
   slot[t].reset();
 }
 if (iie.iiemode)
 {
  iie.rdaux=iie.wraux=iie.lcaux=0;
  iie.cxoverride=0;
  iie.nogx=iie.split=iie.page2=iie.hires=0;
  iie.store80=iie.dblgrx=iie.txt80=iie.mostxt=0;
  iie.an3=0x00;
  iie.c3override=0x00;
 }
 cpureset();
}

uint8_t MHz;

LARGE_INTEGER currenttime,throttlerate;
long long wantedtime;

/* Imprecise */
static void throttle (void)
{
 QueryPerformanceCounter(&currenttime);
 QueryPerformanceFrequency(&throttlerate);
 while (currenttime.QuadPart<wantedtime)
 {
  QueryPerformanceCounter(&currenttime);
  SwitchToThread();
 }
 wantedtime=currenttime.QuadPart+(throttlerate.QuadPart/(15720*MHz));
}

uint8_t ipdump;

/*****************************************************************************/

/*****************************************************************************
 * CPU trace - so I can try to figure out what's wrong with the emulation    *
 *****************************************************************************/

typedef enum {Nothing, Byte, Word, Relative} NumberSize;

typedef struct
{
 char Description[20];
 NumberSize Method;
}
Opcode;

Opcode Ops[256]=
{
 "BRK",Nothing,                 /* $00 */
 "ORA ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "TSB $%02X",Byte,
 "ORA $%02X",Byte,
 "ASL $%02X",Byte,
 "???",Nothing,
 "PHP",Nothing,                 /* $08 */
 "ORA #$%02X",Byte,
 "ASL",Nothing,
 "???",Nothing,
 "TSB $%04X",Word,
 "ORA $%04X",Word,
 "ASL $%04X",Word,
 "???",Nothing,
 "BPL $%04X",Relative,          /* $10 */
 "ORA ($%02X),Y",Byte,
 "ORA ($%02X)",Byte,
 "???",Nothing,
 "TRB $%02X",Byte,
 "ORA $%02X,X",Byte,
 "ASL $%02X,X",Byte,
 "???",Nothing,
 "CLC",Nothing,                 /* $18 */
 "ORA $%04X,Y",Word,
 "INC",Nothing,
 "???",Nothing,
 "TRB $%04X",Word,
 "ORA $%04X,X",Word,
 "ASL $%04X,X",Word,
 "???",Nothing,
 "JSR $%04X",Word,              /* $20 */
 "AND ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "BIT $%02X",Byte,
 "AND $%02X",Byte,
 "ROL $%02X",Byte,
 "???",Nothing,
 "PLP",Nothing,                 /* $28 */
 "AND #$%02X",Byte,
 "ROL",Byte,
 "???",Nothing,
 "BIT $%04X",Word,
 "AND $%04X",Word,
 "ROL $%04X",Word,
 "???",Nothing,
 "BMI $%04X",Relative,          /* $30 */
 "AND ($%02X),Y",Byte,
 "AND ($%02X)",Byte,
 "???",Nothing,
 "BIT $%02X,X",Byte,
 "AND $%02X,X",Byte,
 "ROL $%02X,X",Byte,
 "???",Nothing,
 "SEC",Nothing,                 /* $38 */
 "AND $%04X,Y",Word,
 "DEC",Nothing,
 "???",Nothing,
 "BIT $%04X,X",Word,
 "AND $%04X,X",Word,
 "ROL $%04X,X",Word,
 "???",Nothing,
 "RTI",Nothing,                 /* $40 */
 "EOR ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "???",Nothing,
 "EOR $%02X",Byte,
 "LSR $%02X",Byte,
 "???",Nothing,
 "PHA",Nothing,                 /* $48 */
 "EOR #$%02X",Byte,
 "LSR",Nothing,
 "???",Nothing,
 "JMP $%04X",Word,
 "EOR $%04X",Word,
 "LSR $%04X",Word,
 "???",Nothing,
 "BVC $%04X",Relative,          /* $50 */
 "EOR ($%02X),Y",Byte,
 "EOR ($%02X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "EOR $%02X,X",Byte,
 "LSR $%02X,X",Byte,
 "???",Nothing,
 "CLI",Nothing,                 /* $58 */
 "EOR $%04X,Y",Word,
 "PHY",Nothing,
 "???",Nothing,
 "???",Nothing,
 "EOR $%04X,X",Word,
 "LSR $%04X,X",Word,
 "???",Nothing,
 "RTS",Nothing,                 /* $60 */
 "ADC ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "STZ $%02X",Byte,
 "ADC $%02X",Byte,
 "ROR $%02X",Byte,
 "???",Nothing,
 "PLA",Nothing,                 /* $68 */
 "ADC #$%02X",Byte,
 "ROR",Nothing,
 "???",Nothing,
 "JMP ($%04X)",Word,
 "ADC $%04X",Word,
 "ROR $%04X",Word,
 "???",Nothing,
 "BVS $%04X",Relative,          /* $70 */
 "ADC ($%02X),Y",Byte,
 "ADC ($%02X)",Byte,
 "???",Nothing,
 "STZ $%02X,X",Byte,
 "ADC $%02X,X",Byte,
 "ROR $%02X,X",Byte,
 "???",Nothing,
 "SEI",Nothing,                 /* $78 */
 "ADC $%04X,Y",Word,
 "PLY",Nothing,
 "???",Nothing,
 "JMP ($%04X,X)",Word,
 "ADC $%04X,X",Word,
 "ROR $%04X,X",Word,
 "???",Nothing,
 "BRA $%04X",Relative,          /* $80 */
 "STA ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "STY $%02X",Byte,
 "STA $%02X",Byte,
 "STX $%02X",Byte,
 "???",Nothing,
 "DEY",Nothing,                 /* $88 */
 "BIT #$%02X",Byte,
 "TXA",Nothing,
 "???",Nothing,
 "STY $%04X",Word,
 "STA $%04X",Word,
 "STX $%04X",Word,
 "???",Nothing,
 "BCC $%04X",Relative,          /* $90 */
 "STA ($%02X),Y",Byte,
 "STA ($%02X)",Byte,
 "???",Nothing,
 "STY $%02X,X",Byte,
 "STA $%02X,X",Byte,
 "STX $%02X,Y",Byte,
 "???",Nothing,
 "TYA",Nothing,                 /* $98 */
 "STA $%04X,Y",Word,
 "TXS",Nothing,
 "???",Nothing,
 "STZ $%04X",Word,
 "STA $%04X,X",Word,
 "STZ $%04X,X",Word,
 "???",Nothing,
 "LDY #$%02X",Byte,             /* $A0 */
 "LDA ($%02X,X)",Byte,
 "LDX #$%02X",Byte,
 "???",Nothing,
 "LDY $%02X",Byte,
 "LDA $%02X",Byte,
 "LDX $%02X",Byte,
 "???",Nothing,
 "TAY",Nothing,                 /* $A8 */
 "LDA #$%02X",Byte,
 "TAX",Nothing,
 "???",Nothing,
 "LDY $%04X",Word,
 "LDA $%04X",Word,
 "LDX $%04X",Word,
 "???",Nothing,
 "BCS $%04X",Relative,          /* $B0 */
 "LDA ($%02X),Y",Byte,
 "LDA ($%02X)",Byte,
 "???",Nothing,
 "LDY $%02X,X",Byte,
 "LDA $%02X,X",Byte,
 "LDX $%02X,Y",Byte,
 "???",Nothing,
 "CLV",Nothing,                 /* $B8 */
 "LDA $%04X,Y",Word,
 "TSX",Nothing,
 "???",Nothing,
 "LDY $%04X,X",Word,
 "LDA $%04X,X",Word,
 "LDX $%04X,Y",Word,
 "???",Nothing,
 "CPY #$%02X",Byte,             /* $C0 */
 "CMP ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "CPY $%02X",Byte,
 "CMP $%02X",Byte,
 "DEC $%02X",Byte,
 "???",Nothing,
 "INY",Nothing,                 /* $C8 */
 "CMP #$%02X",Byte,
 "DEX",Nothing,
 "???",Nothing,
 "CPY $%04X",Word,
 "CMP $%04X",Word,
 "DEC $%04X",Word,
 "???",Nothing,
 "BNE $%04X",Relative,          /* $D0 */
 "CMP ($%02X),Y",Byte,
 "CMP ($%02X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "CMP $%02X,X",Byte,
 "DEC $%02X,X",Byte,
 "???",Nothing,
 "CLD",Nothing,                 /* $D8 */
 "CMP $%04X,Y",Word,
 "PHX",Nothing,
 "???",Nothing,
 "???",Nothing,
 "CMP $%04X,X",Word,
 "DEC $%04X,X",Word,
 "???",Nothing,
 "CPX #$%02X",Byte,             /* $E0 */
 "SBC ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "CPX $%02X",Byte,
 "SBC $%02X",Byte,
 "INC $%02X",Byte,
 "???",Nothing,
 "INX",Nothing,                 /* $E8 */
 "SBC #$%02X",Byte,
 "NOP",Nothing,
 "???",Nothing,
 "CPX $%04X",Word,
 "SBC $%04X",Word,
 "INC $%04X",Word,
 "???",Nothing,
 "BEQ $%04X",Relative,          /* $F0 */
 "SBC ($%02X),Y",Byte,
 "SBC ($%02X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "SBC $%02X,X",Byte,
 "INC $%02X,X",Byte,
 "???",Nothing,
 "SED",Nothing,                 /* $F8 */
 "SBC $%04X,Y",Word,
 "PLX",Nothing,
 "???",Nothing,
 "???",Nothing,
 "SBC $%04X,X",Word,
 "INC $%04X,X",Word,
 "???",Nothing
};

int dasm(char *output, unsigned short int address)
{
 /* Addresses used by branch ops are relative to address+2, signed char */
 unsigned int num;
 signed char Rel;
 unsigned int p1,p2;
 unsigned int op;
 char Desc[128];

 /* We must keep the compiler from promoting signed values, so often we
    explicitly cast to unsigned char. */
 op=cpuread(address);

 if (Ops[op].Method==Nothing) /* Just the opcode */
 {
  sprintf (output,"%04X:  %02X            %s",address,op,Ops[op].Description);
  return address+1;
 }
 if (Ops[op].Method==Byte) /* Two byte opcode */
 {
  sprintf (Desc,Ops[op].Description,(unsigned char)(p1=cpuread(address+1)));
  sprintf (output,"%04X:  %02X%02X          %s",address,op,(unsigned char)p1,Desc);
  return address+2;
 }
 if (Ops[op].Method==Word) /* Three byte opcode */
 {
  p1=(unsigned char)cpuread(address+1);
  p2=(unsigned char)cpuread(address+2);
  num=p1+(p2<<8); /* Construct low-endian number */
  sprintf (Desc,Ops[op].Description,num);
  sprintf (output,"%04X:  %02X%02X%02X        %s",address,op,p1,p2,Desc);
  return address+3;
 }
 if (Ops[op].Method==Relative) /* Two byte opcode */
 {
  /* This is a little tricky... The relative address is a signed char. */
  Rel=cpuread(address+1);
  num=(address+2);
  num+=Rel; /* this CAN subtract; remember a+(-b) is the same as a-b */
  sprintf (Desc,Ops[op].Description,(unsigned int)num);
  sprintf (output,"%04X:  %02X%02X          %s",address,op,(unsigned char)Rel,Desc);
  return address+2;
 }
 /* error : should never reach this point but placate compilers */
 return -1;
}

/*****************************************************************************/

void dropline(void)
{
 uint16_t ip;
 char buf[80];
 
 ip=cpugetpc();
 printf("%04X- A=%02X X=%02X Y=%02X S=%02X P=%02X\n",ip,cpugeta(),cpugetx(),cpugety(),cpugets(),cpugetp());
 dasm(buf,ip);
 printf("%s\n",buf);
}

int8_t cpuloop(uint32_t cycles)
{
 int x;

#ifdef PROFILE
 printf (".");
#endif
 
 if (ipdump) dropline();
 
 ulscanline++;
 if (!fast6on) throttle();
 while (ulscanline>261)
 {
  ulscanline-=262;
 }
 
 if (sndon&&(!fast6on))
 {
  sndbuf[sndbufptr++]=polarity*amplitude;
  if (sndbufptr>=(157200))
  {
   sndbufptr=0;
  }
  else if (sndbufptr==524)
  {
   waveOutWrite(hwaveout,&hdr,sizeof(hdr));
  }
 }
 
 for (x=iie.iiemode?1:0;x<8;x++)
 {
  if (slot[x].dll) slot[x].loop(cpugetcycles());
 }
 x=ttyloop();
 if (x==6502)
 {
  ipdump=!ipdump;
  cpusetrate(ipdump?1:65*cpumultiplier);
  printf ("--\n");
 }
 if (x==-1) return 3;
 if (x==1) reset_all();
 if ((x>=10)&&(x<18))
 {
  if (slot[x-10].dll)
  {
   slot[x-10].config();
  }
 }
 return 0;
}

int slotassign (int slotno, char *module)
{
 int e;
 
 if ((slotno<0)||(slotno>7))
 {
  fprintf (stderr, ESTR_BADSLOT, slotno);
  return -1;
 }
 
 slot[slotno].dll=LoadLibrary(module);
 if (!(slot[slotno].dll))
 {
  fprintf(stderr, ESTR_OPEN, module);
  return -1;
 }

 if ((!(slot[slotno].mod_type=(void*)GetProcAddress(slot[slotno].dll,MOD_TYPE)))||
     (!(slot[slotno].mod_name=(void*)GetProcAddress(slot[slotno].dll,MOD_NAME)))||
     (!(slot[slotno].mod_pref=(void*)GetProcAddress(slot[slotno].dll,MOD_PREF)))||
     (!(slot[slotno].mod_vers=(void*)GetProcAddress(slot[slotno].dll,MOD_VERS)))||
     (!(slot[slotno].mod_copr=(void*)GetProcAddress(slot[slotno].dll,MOD_COPR)))) notamod(module);
 
 if ((*(slot[slotno].mod_type))!=0x544F4C53)
 {
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  fprintf (stderr, ESTR_ISNOT, module);
  fprintf (stderr, ESTR_NOTSLOT);
  return -1;
 }
 
 SFUNCTINO(slotno, preload_name, "preload_name");
 SFUNCTINO(slotno, preload_size, "preload_size");
 SFUNCTINO(slotno, preload_ptr,  "preload_ptr");

 SFUNCTINO(slotno, write16, "write16");
 SFUNCTINO(slotno, write256, "write256");
 SFUNCTINO(slotno, write2048, "write2048");

 SFUNCTINO(slotno, read16, "read16");
 SFUNCTINO(slotno, read256, "read256");
 SFUNCTINO(slotno, read2048, "read2048");

 SFUNCTINO(slotno, override_rd, "override_rd");
 SFUNCTINO(slotno, override_wr, "override_wr");
 
 SFUNCTINO(slotno, reset, "reset");
 SFUNCTINO(slotno, loop, "loop");
 SFUNCTINO(slotno, init, "init");
 SFUNCTINO(slotno, deinit, "deinit");

 SFUNCTINO(slotno, config, "config");

 if (*(slot[slotno].preload_ptr))
 {
  if (bload(*(slot[slotno].preload_name),*(slot[slotno].preload_size),*(slot[slotno].preload_ptr),0))
  {
   FreeLibrary(slot[slotno].dll);
   slot[slotno].dll=0;
   return -1;
  }
 }
 
 e=slot[slotno].init(slotno);
 if (e==6502) /* plugin needs dma */
 {
  SFUNCTINO(slotno, dmainit, "dmainit");
  slot[slotno].dmainit(cpuwrite, cpuread);
 }
 else if (e)
 {
  fprintf (stderr, ESTR_INITMOD, module);
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  return -1;
 }
 
 printf (ESTR_SLOT, slotno);
 printf (ESTR_IS,*(slot[slotno].mod_name),*(slot[slotno].mod_vers),*(slot[slotno].mod_copr));
 return 0;
}

int WINAPI WinMain(HINSTANCE ghinstance, HINSTANCE dummy, LPSTR cmdline, int nCmdShow)
{
 char inifile[1024];
 char modbuf[256], slotstr[6];
 char ttymod[256], cpumod[256];
 int s,l;
 int gotrom;
 
 hinstance=ghinstance;
 
 gotrom=0;
 
 getcwd(inifile,1010);
 strcat(inifile,"\\modapple.ini");
 
 for (l=0; l<8; l++) slot[l].dll=0;
 
 printf ("%s\n", verstr);
 
 l=*ttymod=*cpumod=0;
 
 l=GetPrivateProfileString(INI_SYSTEM, INI_CONSOLE, 0, modbuf, 256, inifile);
 if (l)
 {
  strcpy(ttymod, modbuf);
 }
 
 l=GetPrivateProfileString(INI_SYSTEM, INI_CPU, 0, modbuf, 256, inifile);
 if (l)
 {
  strcpy(cpumod, modbuf);
 }

 cpumultiplier=1;
 l=GetPrivateProfileString(INI_SYSTEM, INI_MULT, 0, modbuf, 256, inifile);
 if (l)
 {
  int32_t m;
  
  m=atol(modbuf);
  if ((m>0)&&(m<1000)) cpumultiplier=m;
 }

 for (s=0; s<8; s++)
 {
  sprintf (slotstr, "Slot%d", s);
  l=GetPrivateProfileString(INI_SLOTCARDS, slotstr, 0, modbuf, 256, inifile);
  if (l)
  {
   if (slotassign(s,modbuf))
    fprintf (stderr, ESTR_LINK, modbuf, s);
  }
 }
 
 if (!*ttymod)
 {
  fprintf (stderr, ESTR_NOTTY);
  return -1;
 }
 
 if (!*cpumod)
 {
  fprintf (stderr, ESTR_NOCPU);
  return -1;
 }
 
 mod_tty=0;
 mod_cpu=0;
 
 mod_tty=LoadLibrary(ttymod);
 if (!mod_tty)
 {
  fprintf(stderr, ESTR_OPEN, ttymod);
  return -1;
 }
 if ((!(tty_mod_type=(void*)GetProcAddress(mod_tty,MOD_TYPE)))||
     (!(tty_mod_name=(void*)GetProcAddress(mod_tty,MOD_NAME)))||
     (!(tty_mod_pref=(void*)GetProcAddress(mod_tty,MOD_PREF)))||
     (!(tty_mod_vers=(void*)GetProcAddress(mod_tty,MOD_VERS)))||
     (!(tty_mod_copr=(void*)GetProcAddress(mod_tty,MOD_COPR)))) notamod(ttymod);
 
 if ((*tty_mod_type)!=0x50595454)
 {
  canlibs();
  fprintf (stderr, ESTR_ISNOT, ttymod);
  fprintf (stderr, ESTR_NOTTTY);
  return -1;
 }
 
 TFUNCTINO(ttykey,"ttykey");
 TFUNCTINO(ttywrite,"ttywrite");
 TFUNCTINO(ttyiie,"ttyiie");
 TFUNCTINO(ttyout,"ttyout");
 TFUNCTINO(ttyin,"ttyin");
 TFUNCTINO(ttydeinit,"ttydeinit");
 TFUNCTINO(ttyinit,"ttyinit");
 TFUNCTINO(ttyloop,"ttyloop");
 
 mod_cpu=LoadLibrary(cpumod);
 if (!mod_cpu)
 {
  fprintf(stderr, ESTR_OPEN, cpumod);
  return -1;
 }
 if ((!(cpu_mod_type=(void*)GetProcAddress(mod_cpu,MOD_TYPE)))||
     (!(cpu_mod_name=(void*)GetProcAddress(mod_cpu,MOD_NAME)))||
     (!(cpu_mod_pref=(void*)GetProcAddress(mod_cpu,MOD_PREF)))||
     (!(cpu_mod_vers=(void*)GetProcAddress(mod_cpu,MOD_VERS)))||
     (!(cpu_mod_copr=(void*)GetProcAddress(mod_cpu,MOD_COPR)))) notamod(cpumod);
 
 if ((*cpu_mod_type)!=0x32303536)
 {
  canlibs();
  fprintf (stderr, ESTR_ISNOT, cpumod);
  fprintf (stderr, ESTR_NOTCPU);
  return -1;
 }
 
 CFUNCTINO(cpureset,"cpureset");
 CFUNCTINO(cpurun,"cpurun");
 CFUNCTINO(cpuinit,"cpuinit");
 CFUNCTINO(cpuseta,"cpuseta");
 CFUNCTINO(cpusetx,"cpusetx");
 CFUNCTINO(cpusety,"cpusety");
 CFUNCTINO(cpusets,"cpusets");
 CFUNCTINO(cpusetp,"cpusetp");
 CFUNCTINO(cpusetpc,"cpusetpc");
 CFUNCTINO(cpusetrate,"cpusetrate");
 CFUNCTINO(cpugeta,"cpugeta");
 CFUNCTINO(cpugetx,"cpugetx");
 CFUNCTINO(cpugety,"cpugety");
 CFUNCTINO(cpugets,"cpugets");
 CFUNCTINO(cpugetp,"cpugetp");
 CFUNCTINO(cpugetpc,"cpugetpc");
 CFUNCTINO(cpugetrate,"cpugetrate");
 CFUNCTINO(cpugetcycles,"cpugetcycles");

 l=GetPrivateProfileString(INI_SYSTEM, INI_BIOS, INI_DEF_ROM, modbuf, 256, inifile);

 iie.iiemode=0;
 iie.ram=ram;
 
 if (!(bload(modbuf,12288,rom,1))) 
  gotrom=1;
 else
  if (!(bload(modbuf,16384,rom,0)))
  {
   iie.iiemode=1;
   gotrom=1;
  }
 if (!gotrom) fprintf (stderr, ESTR_NOBIOS);
 
 if (iie.iiemode)
 {
  uint32_t crc;
  
  crc=crc32(0,rom,16384);
  switch (crc)
  {
   case 0x281592E5: /* original ROM */
    iie.enh=0;
    break;
   case 0x1D70B193: /* enhanced ROM */
   case 0x7FC03D24: /* EDM */
    iie.enh=1;
    break;
   default:
    iie.enh=0;
  }
 }

 printf (ESTR_CPU);
 printf (ESTR_IS, *cpu_mod_name, *cpu_mod_vers, *cpu_mod_copr);
 printf (ESTR_TTY);
 printf (ESTR_IS, *tty_mod_name, *tty_mod_vers, *tty_mod_copr);
 printf (ESTR_MULT,cpumultiplier);

 holdc8=0;
 
#ifndef ENABLE_SOUND
 sndon=0;
#else
 if (openwav())
 {
  sndon=0;
  fprintf (stderr, ESTR_SND1);
 }
 else
 {
  hdr.dwBufferLength=(157200);
  hdr.dwBytesRecorded=0;
  hdr.dwLoops=0;
  hdr.dwFlags=0;
  hdr.lpData=sndbuf;
  if (waveOutPrepareHeader(hwaveout,&hdr,sizeof(hdr)))
  {
   waveOutClose(hwaveout);
   sndon=0;
   fprintf (stderr, ESTR_SND2);
  }
  else
  {
   printf (ESTR_SND3);
   sndon=1;
  }
 }
 polarity=0;
 amplitude=63;
#endif
 
 joyinit();
 
 cpuinit(cpuwrite, cpuread, cpuloop);

 cpureset();
 cpusetrate(65*cpumultiplier);
 
 MHz=cpumultiplier;
 wantedtime=0;
 fast6on=0;
 fast6=1;
 ipdump=0;
 
 *ttykey=nCmdShow;
 ttyinit(&iie);
 reset_all();
 cpurun();
 ttydeinit();
 for (l=7; l>=0; l--) if (slot[l].dll) slot[l].deinit();
 
 canlibs();
 
 return 0;
}
